#include <Arduino.h>
#include <ArduinoLog.h>

#include "config.h"
#include "pins.h"
#include "nfc.h"
#include "otaproxy.h"
#include "helpers.h"
#include "Desfire.h"

#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>

#include <WiFiClient.h>
#include <PubSubClient.h>
#include <SPI.h>
#include <SSD1306Wire.h>

WiFiClient espClient;
//Config_Data config;

PubSubClient* mqtt;
NFC* nfc;
OTAProxy* ota;

unsigned long otatimeout = 3000;
unsigned long lastotatime;

void setup_wifi() 
{
    delay(10);
    Serial.println("Connecting Wifi ...");

    WiFi.mode(WIFI_STA);
    WiFi.begin(WLAN_SSID, WLAN_PASS);

    while (WiFi.status() != WL_CONNECTED)
    {
        delay(500);
    }

    randomSeed(micros());
    Serial.println("WiFi connected");
}

void reconnect() 
{
    while (!mqtt->connected())
    {
        String clientId = "FabReader_";
        clientId += String(FABREADERID);

        Serial.println("Connecting MQTT ...");
        if (mqtt->connect(clientId.c_str()))
        {
            Serial.println("MQTT connected");
            char id[6] = "00000";
            sprintf(id, "%05d", FABREADERID);
            mqtt->publish("fabreader", id);

            char topic_requestOTA[] = "fabreader/00000/requestOTA";
            sprintf(topic_requestOTA, "fabreader/%05d/requestOTA", FABREADERID);
            mqtt->subscribe(topic_requestOTA);

            char topic_stopOTA[] = "fabreader/00000/stopOTA";
            sprintf(topic_stopOTA, "fabreader/%05d/stopOTA", FABREADERID);
            mqtt->subscribe(topic_stopOTA);
        } else {
            Serial.print("failed, rc=");
            Serial.print(mqtt->state());
            Serial.println(" try again in 5 seconds");
            delay(5000);
        }
    }
}

void callback(char* topic, byte* payload, unsigned int length) 
{
    Serial.println("Receive Message");
    Serial.println(topic);
    if(ota->hasActiveOTA())
    {
        ota->continueOTA(topic, payload, length);
    }
}

void setup()
{
    Serial.begin(115200);
    Serial.print("\n\n\n");
    Serial.println("Booting ...");
    
    pinMode(PIN_BUZZER, OUTPUT);
    pinMode(PIN_BUTTON, INPUT);
    pinMode(PIN_ETH_SPI_SS, OUTPUT);
    digitalWrite(PIN_ETH_SPI_SS, HIGH);

    Serial.println("Connecting NFC ...");
    nfc = new NFC(PIN_RFID_SPI_SS, PIN_RFID_RST);
    if(!(nfc->rfid->PCD_PerformSelfTest()))
    {
        Serial.println("NFC Test failed");
    }
    Serial.println("NFC connected");

    setup_wifi();

    mqtt = new PubSubClient(espClient);
    mqtt->setServer(MQTT_BROKER, 1883);
    mqtt->setCallback(callback);

    ota = new OTAProxy(mqtt, nfc, FABREADERID);
}

void loop()
{    
    if (!mqtt->connected())
    {
        reconnect();
    }
    mqtt->loop();

    if(!ota->hasActiveOTA())
    {
        if(nfc->checkforCard())
        {
            Serial.println("Card detected");
            if(nfc->connecttoCard())
            {
                Serial.println("Card connected");
                lastotatime = millis();
                ota->startOTA();
            }
        }
    }
    if(ota->hasActiveOTA())
    {
        if(millis() - lastotatime > otatimeout)
        {
            ota->cancelOTA();
        }
    }
}